/* Contextual Exceptions classes for throwing exceptions that know where they
   came from.
   Copyright (C) 2002 Aaron Bentley

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */

/**
 * \file context.cpp
 * \brief Implementation of Context, ContextStack, ContException, for context-
 * aware exceptions
 */
#include "context.h"
#include <iostream>

using std::cout;

using namespace context;

/**
 * \brief Constructs an object of this type and adds it to the ContextStack
 * \param argd A string description of this context
 */
Context::Context(const string &argd)
   :
description(argd)
{
   ContextStack::getInstanceOf().push(*this);
}


/**
 * \brief Destructs this object and removes it from the ContextStack
 */
Context::~Context()
{
   ContextStack::getInstanceOf().pop(*this);
}


/**
 * \brief Returns a description of this object
 */
const std::string &Context::getDescription() const
{
   return description;
}


/**
 * \brief Prints a vector of strings as a stack (FILO)
 * \param list The string stack
 * \param start The start position
 * \param end The end position
 */
void Context::print(const vector<string> &list, size_t start=0, 
   size_t end=0)
{
   if (start>=list.size()) return;
   end=(end>start)?end:list.size();

   for (size_t i=end; i>start; --i)
   {
      std::cout<<list[i-1]<<"\n";
   }
}

/*****************************************************************************
 * ContextStack begins
 *****************************************************************************/

//Initialize to a sentinel value
ContextStack::ContextStack *ContextStack::instance=NULL;


/**
 * \brief Adds a Context to the end of the stack
 * \param cont The Context to add
 */
void ContextStack::push(Context &cont)
{
   stack.push_back(&cont);
}


/**
 * \brief Removes a Context from the end of the stack
 */
void ContextStack::pop()
{
   stack.pop_back();
}


/**
 * \brief Removes a Context from the end of the stack
 *
 * Prints interesting errors if the wrong Context is removed
 * \param cx The Context to remove
 */
void ContextStack::pop(Context &cx)
{

   if (stack.size()==0) 
   {
      cout<<"Attempt to pop empty stack\n";
      return;
   }
   if (&cx!=stack[stack.size()-1])
   {
      cout <<"Attempt to pop wrong context:\n";
      cout <<"\""<<cx.getDescription()<<"\" (cx) should be\n";
      cout <<"\""<<stack[stack.size()-1]->getDescription()<<"\" (*stack[";
      cout <<stack.size()<<"])\n";
      return;
   }
   pop();
}


/**
 * \brief Produces a vector of strings representing the current ContextStack
 * \return An auto_ptr to the vector of strings created
 */
auto_ptr< vector<string> > ContextStack::makeSVector() const
{
   auto_ptr< vector<string> > myvector(new vector<string>);
   if (myvector.get()==NULL) return myvector;
   for (size_t i=0; i<stack.size(); ++i)
   {
      myvector->push_back(stack[i]->getDescription());
   }

   return myvector;
}


/**
 * \brief Prints the current context
 */
void ContextStack::print()
{
   auto_ptr< vector<string> > cxv(getInstanceOf().makeSVector());
   Context::print(*cxv);
}

/******************************************************************************
 * ContException begins
 ******************************************************************************/

/**
 * \brief Constructs the ConstException, using the current ContextStack
 */
ContException::ContException()
   :
thrContext(ContextStack::getInstanceOf().makeSVector())
{}

/**
 * \brief Constructs the ConstExceptions, using the current ContextStack
 * \param argmsg The message this object should represent
 */
ContException::ContException(const string &argmsg)
   :
thrContext(ContextStack::getInstanceOf().makeSVector()), msg(argmsg)
{}

/**
 * \brief Prints the message, its (partial) context, and the current context
 * (which should be a subset of the context it was thrown from)
 */
void ContException::print()
{
   auto_ptr <vector <string> > 
      prContext(ContextStack::getInstanceOf().makeSVector());
   // Delay setting context until the string vector is produced, since
   // ConstException::print does not belong in the output
   Context cx("ContException::print");
   cout << msg << "\n";
   cout << "Thrown in context:\n";
   Context::print(*thrContext, prContext->size(), thrContext->size());
   cout << "Printed in context:\n";
   Context::print(*prContext);
}

