#include <string>
#include <map>
#include <vector>

   using std::string;
/**
 * \brief This class (and its subclasses) can be used to set and display
 * variables.
 *
 * They are designed for temporary use, so they take references rather than
 * containing their variables
 */
class EditVar
{
   protected:
/// The name to display in UI contexts (may contain spaces, need not be unique)
   string displayname;
/// The name to use internally (may not contain spaces, should be unique)  
   string varname;

   public:
   virtual ~EditVar()
   { ; }
/**
 * \brief Assigns the referenced variable from a string
 * \param input The string representation of the variable
 * \return This EditVar
 */
   virtual EditVar &operator=(string const &input)=0;

/**
 * \brief Assigns the referenced variable from a string map
 *
 * The varname is used to look up the value in the map
 * \param input The map that may contain a pair for the varname
 * \return This EditVar
 */
   virtual EditVar &operator=(std::map<string, string> &input)
   {
      operator=(input[varname]);
      return *this;
   }

/**
 * \brief Writes the referenced variable out as an HTML FORM variable, in one
 * line
 * 
 * Writes the display name first, then the appriopriate FORM elements
 * \param html The string to output to
 */
   virtual void htmlAppend(string &html) const
   {
      html+=getdisplay();
      html+=' ';
      formEl(html);
   }
/**
 * \brief Writes the referenced variable out as an HTML FORM variable, as a
 * line in a table
 * 
 * Writes the display name first, then the appriopriate FORM elements
 * \param html The string to output to
 */
   virtual void htmlTable(string &html) const;

/**
 * \brief Writes the referenced variable out as an HTML FORM element
 * \param html The string to output to
 */
   virtual void formEl(string &html) const=0;

/**
 * \brief Determines the value of the variable as a string
 */
   virtual string stringVal();

/**
 * \brief Determines the element's display name
 */
   const string &getdisplay() const
   {
      return displayname;
   }
   //virtual void addXml(xml::Xtree &parent) const=0;
   /*virtual EditVar &operator=(xml::Xtree const &parent)
   {
      string const *str=parent.getVal(varname);
      if (str!=NULL) operator=(*str);
      return *this;
   }*/
};

/**
 * \brief A list of (pointers to) EditVars
 *
 * EditVars that are added to the list are deleted by the destructor
 */
class EditVarList: public std::vector<EditVar *>
{  
   public:
/**
 * \brief Assigns all the variables referenced by EditVars in the list from the
 * input
 * \param input The map that contains values for the EditVars
 */
   void operator=(std::map<string, string> &input);
   
   //void operator=(xml::Xtree const &parent);

/**
 * \brief Outputs all the EditVars as rows in an HTML table
 * \param html The string to write the HTML to
 */
   void htmlTable(string &html);

/**
 * \brief Adds XML representations of all EditVars to the given Xtree
 * \param xml The XML tree to add the EditVars to
 */
   //void addXml(xml::Xtree &xml);
};

/**
 * \brief Base class for objects that can be set, displayed and written
 * using EditVar
 */
class Editable
{
protected:
   /**
    * \brief This method produces a list of EditVars that wrap the object's real
    * member variables
    * \note This method must be implemented in all subclasses
    */
   virtual void getEditVars(EditVarList &)=0;

public:
   virtual ~Editable()
   {;}

   /**
    * \brief Assigns the variables from a map using the EditVarList 
    * \param input The map containing values for the variables
    */
   virtual Editable &operator=(std::map<string, string> &input);

   /**
    * \brief Adds HTML for the variables to a string using the EditVarList
    * \param html The string to append HTML to
    * \return A reference to this
    */
   virtual void htmlTable(string &html) const;

/**
 * \brief Adds XML for the variables to an Xtree
 * \param xml The Xtree to add variables to
 */
//   virtual void addXml(xml::Xtree &xml) const;
/**
 * \brief Initializes settings for the variables from an Xtree
 * \param xml The Xtree to get variables from
 */
//   virtual Editable &operator=(xml::Xtree const &xml);
};



/**
 * \brief An EditVar that uses a boolean and represents it as a checkbox
 */
class EditBool: public EditVar
{
   private:
   bool &mybool;
   public:
/**
 * \brief Constructor that sets up the reference and names
 * \param argbool The boolean that this EditVar should use
 * \param argvarname The variable name that this Editvar should use
 * \param argdisplayname The name that should be displayed for this EditVar
 */
   EditBool(bool &argbool, const string &argvarname, 
      const string &argdisplayname):
      mybool(argbool)
   {
      displayname=argdisplayname;
      varname=argvarname;
   }

   EditVar &operator=(string const &input);
/**
 * \brief Assigns the referenced variable from a boolean
 */
   EditBool &operator=(bool newval)
   {
      mybool=newval;
      return *this;
   }
/**
 * \brief Assigns the referenced variable from a char array
 */
   EditVar &operator=(char const input[]);
   void formEl(string &html) const;
/**
 * \brief Returns the value of the referenced variable
 */
   bool getBool() const
   {
      return mybool;
   }
   /*void addXml(xml::Xtree &parent) const
   {
      string value;
      value=(mybool)?"true":"false";
      parent.addChild(xml::newTag(varname, value));
   }*/
};

/**
 * \brief An EditVar that uses a string and represents it as a field
 */
class EditFieldString: public EditVar
{
   private:
   string &mystring;
   public:
/**
 * \brief Constructor that sets up the reference and names
 * \param argstring The string that this EditVar should use
 * \param argvarname The variable name that this Editvar should use
 * \param argdisplayname The name that should be displayed for this EditVar
 */
   EditFieldString(string &argstring, const string &argvarname, 
      const string &argdisplayname):
      mystring(argstring)
   {
      displayname=argdisplayname;
      varname=argvarname;
   }

   EditVar &operator=(string const &input);
   
   void formEl(string &html) const;
   /*void addXml(xml::Xtree &parent) const
   {
      parent.addChild(xml::newTag(varname, mystring));
   }*/
};

/**
 * \brief An EditVar that uses a string and represents it as a text box
 */
class EditBoxString: public EditVar
{
   private:
   string &mystring;
   public:
/**
 * \brief Constructor that sets up the reference and names
 * \param argstring The string that this EditVar should use
 * \param argvarname The variable name that this Editvar should use
 * \param argdisplayname The name that should be displayed for this EditVar
 */
   EditBoxString(string &argstring, const string &argvarname, 
      const string &argdisplayname):
      mystring(argstring)
   {
      displayname=argdisplayname;
      varname=argvarname;
   }

   EditVar &operator=(string const &input);
   void htmlTable(string &html) const;
   void formEl(string &html) const;
   /*void addXml(xml::Xtree &parent) const
   {
      parent.addChild(xml::newTag(varname, mystring));
   }*/
};

/**
 * \brief An EditVar that uses an int and represents it as a field
 */
class EditInt: public EditVar
{
   private:
   int &myint;
   public:
/**
 * \brief Constructor that sets up the reference and names
 * \param argstring The string that this EditVar should use
 * \param argvarname The variable name that this Editvar should use
 * \param argdisplayname The name that should be displayed for this EditVar
 */
   EditInt(int &argint, const string &argvarname, 
      const string &argdisplayname):
      myint(argint)
   {
      displayname=argdisplayname;
      varname=argvarname;
   }

   EditVar &operator=(string const &input);
   
   void formEl(string &html) const;
   /*void addXml(xml::Xtree &parent) const
   {
      parent.addChild(xml::newTag(varname, "pi"));
   }*/
};

/*class Gig:public Editable
{
   private:
   string venueID, time, description;
   protected:
   EditVarList *getEditVars();
   public:
   using Editable::operator=;
};*/

string &operator+=(string &html, const EditVar &var);

/**
 * \brief Appends all the variables in the EditVarList to the string as an HTML
 * table 
 */
string &operator+=(string &html, EditVarList &vars);

string &operator+=(string &html, Editable &obj);

/**
 * \brief Adds all the variables in the Editable to the Xtree as XML
 */
/*xml::Xtree &operator+=(xml::Xtree &xml, Editable &obj)
{
   obj.addXml(xml);
   return xml;
}*/
